// rewritten by Zheng Gengbin and Shi Lixia
// perform:
// static communication detection

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#else
#include <strings.h>
extern "C" void exit(...);
extern "C" int abs(...);
#endif

#include "compilerApi.h"
#include "strategy.h"

#include "anls_att.h"
#include "detect_comm.h"


extern void analyse(SgStatement * s);

extern int pcrc_detect_communication(detect_DIM* dad_x, detect_DIM* dad_y,
                              int* l, int* u, int* stride, int *a0, int *b0,
                              int* a1, int* b1, int *amount);

extern int shift_is_ok(SgStatement *s, SgSymbol *arraysymb);
extern Shadow_def * get_defined_shadow(SgSymbol *);
extern int getreflevel(SgSymbol * index, SgExpression * expr);
extern SgExpression * getlinerexp(int i, SgSymbol * symb, SgExpression * expr);
extern void change_scalar_expression(SgStatement * , SgExpression * );
extern SgExpression *coefficient(SgExpression *expr, SgSymbol *s);

// find if we already have such an expression inserted
int find_same_as(SgStatement *s , SgExpression *expr)
{
 for (int i=1; i<s->numberOfAttributes(Forall_Ref); i++) {
    	forall_reference * rightref = (forall_reference *) (s->attributeValue(i,Forall_Ref));
	if ( (*rightref->getexpression()) == *expr )
	{
//printf("same as :%d\n",i);
		return i;
	}
 }
 return -1;
}

void get_forall_expression(SgStatement * s, SgExpression * expr, Position side)
{
  if(expr) {
    // already in node code status, no need to analyse.
    if (expr->haveAttribute(NODE_CODE_BIT)) return;

    switch(expr->variant()) {
    case BOOL_VAL:
    case INT_VAL:
    case FLOAT_VAL:
    case DOUBLE_VAL:
      break;

    case COMPLEX_VAL: {			// zheng.gb
	// add cmplx()
      //SgSymbol * sym=s->getSymbolInCurrentScope(NULL, nameOfType(expr->operand(1)->type()), "cmplx");
      SgSymbol * sym=s->getSymbolInCurrentScope("cmplx");
      SgFunctionRefExp * funExp = new SgFunctionRefExp(*sym);
      funExp->addArg(*expr->lhs());
      funExp->addArg(*expr->rhs());
      expr->replace(*funExp);
      get_forall_expression(s, funExp->arg(0), side);
      get_forall_expression(s, funExp->arg(1), side);
      break;
    }
    
    case VAR_REF: {
      forall_reference * fref= new forall_reference(side);
      fref->setExpr(expr);
      fref->sameAs = find_same_as(s,expr);
      s->addAttribute(Forall_Ref,(void *)fref, sizeof(forall_reference));
      break;
    };

    case ARRAY_REF: {
      forall_reference * fref= new forall_reference(side);
      fref->setExpr(expr);
      fref->sameAs = find_same_as(s,expr);
      s->addAttribute(Forall_Ref,(void *)fref, sizeof(forall_reference));
      break;
    };

    case MINUS_OP:
    case ADD_OP:  
    case MULT_OP:
    case DIV_OP:
    case EXP_OP:
    case MOD_OP:

    case AND_OP:
    case OR_OP:
    case EQV_OP:
    case NEQV_OP:
      
    case EQ_OP:
    case NOTEQL_OP:
    case LT_OP:
    case LTEQL_OP:
    case GT_OP: 
    case GTEQL_OP: {
      get_forall_expression(s, expr->operand(1), side);
      get_forall_expression(s, expr->operand(2), side);
      break;
    }

    case NOT_OP: {
      get_forall_expression(s, expr->operand(1), side);
      break;
    }

    case FUNC_CALL:
    case INTRIN_FUNC_CALL: {
      SgFunctionRefExp * funExp = (SgFunctionRefExp *) expr;
      Forall_header_def * fhdef = (Forall_header_def *) s->controlParent()->attributeValue(0, FHDER_Def);
      // we use special method to deal the dummy of elementary.
      SgFunctionSymb *funcSym = (SgFunctionSymb*)funExp->symbol();
      if (funcSym->is_elementary_function()==1)
      {
	// look the actual arg as an array
      	for (int i=0; i<funExp->numberOfArgs(); i++) 
      	   get_forall_expression(s, funExp->arg(i), side);
      }
      else
      //  change the indexl reference in argument to pcrc_glob()
      for (int i=0; i<funExp->numberOfArgs(); i++) {
	for (int j=0; j<fhdef->getindexnumber(); j++) {
		SgSymbol *indexsymb=fhdef->getindexsymb(j);
		SgExpression *exp;
		exp = funExp->arg(i)->IsSymbolInExpression(*indexsymb);
		if (exp) change_scalar_expression(s->controlParent(),exp);
	}
      };
      break;
    }

    default:
      if (!silentFlag)
      printf("In get_forall_expression, unexpected expression on line %d, id: %d, varaint: %d\n",s->lineNumber(), expr->id(),expr->variant());

    };
  };
}

void get_forall_reference(SgStatement * s)
{ 
//let all variable or array reference appeared in FORALL be recorded in
//a class variable forall_reference. only the current FORALL statement 
// know this calss variable.

  SgStatement * body = s->childList1(0);
    
  switch(body->variant()) {
  case ASSIGN_STAT: 
    get_forall_expression(body, body->expr(0), left);
    get_forall_expression(body, body->expr(1), right);
    break;

  case CONTROL_END: 
    break;
    
  default:
    printf("Illegal FORALL statement! variant:%d\n",body->variant());
    break;
  };

  Forall_header_def * fhdef=(Forall_header_def *) s->attributeValue(0, FHDER_Def);

  SgExpression * condition = fhdef->getcondition();

  if (condition!=NULL)
    get_forall_expression(body, condition, head);

}

int broadcasting_index_pattern(forall_reference * lhsref, forall_reference * rhsref)
{
  //if left has more index than the right;
  //if less, something wrong;
  // return (1);
  return (0);
}

// return -1, if dim is not partitioned to any dimension of processor.
// return -2, if array is all-collapsed
int get_partitioned_dim(Processor_def * pdef, int dim, SgArrayRefExp * array)
{
   SgSymbol *arraysymb = array->symbol();
   Alignment_def * adef = NULL;
   Distribute_def * ddef = NULL;
   if ( arraysymb->numberOfAttributes(Align_Def) > 0)
   {
     adef = (Alignment_def *)arraysymb->attributeValue(0, Align_Def);
     SgSymbol * tsym = adef->gettosymbol();
     ddef = (Distribute_def *)tsym->attributeValue(0, Dist_Def);
   }
   else {
     ASSERT(arraysymb->numberOfAttributes(Dist_Def));
     ddef = (Distribute_def *) arraysymb->attributeValue(0, Dist_Def);
   }

   int i,j;

   int procdimcounter=0;
   for (i=0; i<ddef->getDim(); i++)
   {
     if (ddef->distCode(i)!=NOP) procdimcounter++;
     if (procdimcounter-1 == dim) break;
   }
   if (i==ddef->getDim()) return -1;
   if (procdimcounter == 0) return -2;		// all collapsed

   for (j=0; j<array->numberOfSubscripts(); j++) {
     int templateDim=adef?adef->getAlignPair(j):j;
     if (templateDim == i)  break;
   }
   if (j==array->numberOfSubscripts()) return -1;
   else return j;

}

// return 1; filled succesully; 0: failed
int fill_DIM(forall_reference * ref, int dim, detect_DIM & dimstruc)
{
  SgSymbol * arraysym = ref->getexpression()->symbol();
  Alignment_def * adef = NULL;
  Distribute_def *ddef = NULL;

  if (arraysym->numberOfAttributes(Align_Def) > 0) {
     adef = (Alignment_def *)arraysym->attributeValue(0, Align_Def);

     ASSERT(adef->getStride(dim)->isInteger());
     ASSERT(adef->getOffset(dim)->isInteger());
     dimstruc.t_stride=adef->getStride(dim)->valueInteger();
     dimstruc.t_offset=adef->getOffset(dim)->valueInteger();

     SgSymbol * templatesym = adef->gettosymbol();
     if (templatesym->numberOfAttributes(Temp_Def))
     {
     Template_def * tdef = (Template_def *)templatesym->attributeValue(0,Temp_Def);
     int temp_dim = adef->getAlignPair(dim);
     if (!tdef->getStart(temp_dim)->isInteger()) return 0;
     dimstruc.t_lb=tdef->getStart(temp_dim)->valueInteger();
     if (!tdef->getEnd(temp_dim)->isInteger()) return 0;
     dimstruc.t_ub=tdef->getEnd(temp_dim)->valueInteger();
     }
     else	// align to a dummy which only have Dist_Def
     {
     SgExpression *lbexp = get_lbound(arraysym,dim);
     if (!lbexp->isInteger()) return 0;		// cannot calculate a value.
     SgExpression *ubexp = get_ubound(arraysym,dim);
     if (!ubexp->isInteger()) return 0;		// cannot calculate a value.
     dimstruc.t_lb=lbexp->valueInteger();
     dimstruc.t_ub=ubexp->valueInteger();
     }

     ddef = (Distribute_def *)templatesym->attributeValue(0, Dist_Def);
  }
  else {	// this is a dummy , so only have Dist_Def
     dimstruc.t_stride=1;
     dimstruc.t_offset=0;
     SgExpression *lbexp = get_lbound(arraysym,dim);
     if (!lbexp->isInteger()) return 0;		// cannot calculate a value.
     SgExpression *ubexp = get_ubound(arraysym,dim);
     if (!ubexp->isInteger()) return 0;		// cannot calculate a value.
     dimstruc.t_lb=lbexp->valueInteger();
     dimstruc.t_ub=ubexp->valueInteger();

     ASSERT(arraysym->numberOfAttributes(Dist_Def));
     ddef = (Distribute_def *)arraysym->attributeValue(0, Dist_Def);
  }

  // modified by Zheng Gengbin
  dim=adef?adef->getAlignPair(dim):dim;

  switch(ddef->distCode(dim)) {
  case NOP:
    dimstruc.distCode=0;
    break;
  case BLOCK:
    dimstruc.distCode=1;
    break;
  case CYCLIC:
    dimstruc.distCode=2;
    break;
  }

  // at present time, we consider distribute is automatically maped to P;

  if (dimstruc.distCode==NOP)
    dimstruc.numProcs=0;
  else {
    int nopnum=0;
    for (int i=0; i<dim; i++) {
      if (ddef->distCode(i)==NOP) 		// bug  : dim=>i
	nopnum++;
    };

    dim-=nopnum;

    SgSymbol * processorsym = ddef->gettosymbol();
    Processor_def * pdef = (Processor_def *) ((processorsym->getAttribute(0, Proc_Def))->getAttributeData());

    SgExpression *numProcs = NULL;
	// have sub processor group
    if (ddef->getProcs())
    {
     for ( int k=0; k<pdef->getDim(); k++)
	numProcs = numProcs * *ddef->getProcSize(k);
    }
    else
        numProcs = &(*pdef->getEnd(dim) - *pdef->getStart(dim) + *new SgValueExp(1) );
    if (!numProcs->isInteger()) return 0;
    dimstruc.numProcs = numProcs->valueInteger();
  };

  return 1;
}


/**********************************************************************/
// begin forall_carray_dependent()

// added by slx 98/1/8

//when return 1, means it cannot calculate subscript, there is flow dependence
//when return 0, calculating subscript is successful

int get_subscript_value(Forall_header_def *fhdef, forall_reference *ref, int i, SgExpression *exp1, SgExpression *exp2, SgExpression *&exp3)
{
  SgArrayRefExp *array=(SgArrayRefExp *)ref->getexpression();

     // get position of index variable that appears in subscript in forall header
   if ( ref->indexNumberOnDim(i) != -1)  // -1 means has scalar dimension
   {
    int index=ref->indexNumberOnDim(i);

      // get index variable's lbound,ubound,step 
    SgExpression *Lb=fhdef->getindextriple(index,0)->copyPtr();
    SgExpression *Ub=fhdef->getindextriple(index,1)->copyPtr();
    SgExpression *Step=fhdef->getindextriple(index,2)?fhdef->getindextriple(index,2)->copyPtr():(new SgValueExp(1));

      //replace index variable in subscript with lbound and ubound of index variable
      //get subscript's lbound(exp1),ubound(exp2),step(exp3) 
    SgSymbol *symb;
    if(ref->getIndex(i,index))          // stride*index+offset
   {
    if (!ref->getstride(i,index)->isInteger())
         return (1);
    if  (ref->getoffset(i,index))
      if  (!ref->getoffset(i,index)->isInteger())
         return (1);
    symb=fhdef->getindexsymb(index);
    exp1->replaceSymbolByExpression(*symb,*Lb);
    exp2->replaceSymbolByExpression(*symb,*Ub);
    exp3=&(*ref->getstride(i,index) * *Step);
   }
  }
   else     //scalar Dim
    {
      if(!array->subscript(i)->isInteger()) return(1); 
      exp3=new SgValueExp(1);
    }

   return (0);
}

int compare_left_right_subscript(int ll, int lu, int ls, int rl, int ru, int rs)
{
/////////////////////////////////////////////////////////////////////
    if ((ll <= lu) && (rl <= ru))            // this means ls>0,rs>0
     {
       // left and right have not common element
          if ((ru<=ll ) || (lu <= rl))              
              return(-1);

       // left < right
          if ((rl >= ll) &&  (rl <= lu))
             return(0);
  
       // left >right
          if ((ll >= rl) &&  (ll <= ru))
         {
           for(int j=ll; j<=ru; j=j+ls)
           {
              if (rs==1)   
                  return (1);
              else 
                 if ((j-rl)% rs==0) 
                     return (1);
           } 
           return (0) ;
          }
      }
//////////////////////////////////////////////////////////////////////  
     if ((ll <= lu) && (rl >= ru))          //this means ls>0, rs<0
     {
        if ((ru >= lu) || (ll >= rl))         
             return(-1);  

         // left >right
        if ((ll <= rl) && (ll >= ru))
        {
          for(int j=ll; j<=rl; j=j+ls)
          {
            if (-rs==1)  
               return (1);
            else
               if ((j-ru)%(-rs)==0)
                   return (1);
          }
          return(0);
        }

        // right > left
        if ((ru <= lu) && (ru >= ll))
        {
          for(int j=ru; j<=lu; j=j-rs)
          { 
             if (ls==1)  
                return (1);
             else
                if ((j-ll)%ls==0)
                   return (1);
          }
          return(0);
        }
     } 
////////////////////////////////////////////////////////////////////
     if ((ll >= lu) && (rl <= ru))         //this means ls<0, rs>0
     {
       if (( ru <= lu) || (ll <=rl))
             return(-1);  

       //  left < right
       if ((rl <= ll) && (rl >=lu))
       {
         for(int j=rl; j<=ll; j=j+rs)
         {
           if (-ls==1)
              return (1);
           else 
              if ((j-lu)%(-ls)==0)
               return (1);
         }
         return(0);
       }

       // left >right
       if ((lu <= ru) && (lu >= rl))
       {
         for (int j=lu; j<=ru; j=j-ls)
         {
           if (rs==1) 
             return (1);
           else
             if ((j-rl)%rs==0)
                return (1);
         }
         return(0);
       } 
     }
//////////////////////////////////////////////////////////////////////////
     if ((ll >= lu) && (rl >= ru))         //this means ls<0, rs<0
     {
       if ((ll <= ru) || (rl <= lu))
             return(-1);

       // left > right 
       if (( lu <= rl) && (lu >= ru))
             return(0);

       // left < right
       if (( ru <= ll) && (ru >= lu))
       {
          for( int j=ru; j<=ll; j=j-rs)
          {
            if(-ls==1)
              return (1);
            else
              if((j-lu)%(-ls)==0)
                 return (1);
          }
          return(0);
       }
     }
     return 1;
}

int forall_carry_dependent(Forall_header_def *fhdef, forall_reference *leftref, forall_reference *rightref)
{
  SgArrayRefExp *larray=(SgArrayRefExp *)leftref->getexpression();
  SgArrayRefExp *rarray=(SgArrayRefExp *)rightref->getexpression();

// get array reference dimension number 
  int numOfSubscript=larray->numberOfSubscripts();
  int i;

  if (leftref->getexpression()->symbol()==rightref->getexpression()->symbol()) 
 { 
   for (i=0; i< numOfSubscript; i++)
   {
    SgExpression *lexp1,*lexp2, *lexp3, *rexp1, *rexp2,*rexp3;
// make copy in order to replace index variable in subscript with lbound and ubound of index variable
    lexp1=larray->subscript(i)->copyPtr();
    lexp2=larray->subscript(i)->copyPtr();
    rexp1=rarray->subscript(i)->copyPtr();
    rexp2=rarray->subscript(i)->copyPtr();

//deal with leftref 
    if (get_subscript_value(fhdef,leftref,i,lexp1,lexp2,lexp3)==1)
        return (1);

// deal with rightref
    if (get_subscript_value(fhdef,rightref,i,rexp1,rexp2,rexp3)==1)
        return (1);

// compare subscript of left and right
// when  ll,lu,ls,rl,ru,rs are integer
   int ll,lu,ls,rl,ru,rs;
   if ((lexp1->valueInteger(ll) != -1) && (lexp2->valueInteger(lu) != -1)
      && (rexp1->valueInteger(rl) != -1) && (rexp2->valueInteger(ru) != -1)
      && (lexp3->valueInteger(ls) != -1) && (rexp3->valueInteger(rs) != -1))
   {
     if (compare_left_right_subscript(ll,lu,ls,rl,ru,rs)==1)
        return (1);
     if (compare_left_right_subscript(ll,lu,ls,rl,ru,rs)==-1)
        return (0);
   }
   else              // one of ll,lu,ls,rl,ru,rs is variable
       return (1);
  }  // end of for
 }  //end of if 

   return (0);

}

/*
int forall_carry_dependent(forall_reference * leftref, 	forall_reference * rightref) {	 
  if (leftref->getexpression()->symbol()==rightref->getexpression()->symbol()) {

    // here we use a simple method, if the correspoding stride and offset is not equal;
    // a dependence is marked;

    for (int i=0; i< MAX_ALIGN_DIM; i++) {
      for (int j=0; j< MAX_INDEX_DIM; j++) {
	if  (leftref->getstride(i, j)!=NULL) {
	  if (rightref->getstride(i, j)!=NULL) {
	    if (leftref->getstride(i,j)->valueInteger()!=rightref->getstride(i,j)->valueInteger())
	      return (1);
	    else if ( leftref->getoffset(i,j) && rightref->getoffset(i,j) )
		if (leftref->getoffset(i,j)->valueInteger()!=rightref->getoffset(i,j)->valueInteger())
	      	return (1);
	  }
	  else
	    return (1);
	}
      }
    }
  }
  return (0);
}
*/

// decide ghostAera for unknow mode
void  set_unknow_ghostarea(SgStatement *s,forall_reference *rhsref)
{
  	forall_reference *lhsref = (forall_reference *)s->childList1(0)->attributeValue(0, Forall_Ref);
        int ishift = 1;

	if ( opt_strategy & RUNTIMECOMM )
	if (rhsref->getCommunication() == UNKNOW)
	{
	  SgSymbol *array = rhsref->getexpression()->symbol();
	  SgArrayType *type = isSgArrayType(array->type());
          ASSERT(type);
          int numDim = type->dimension();
	  int shift_ok = shift_is_ok(s,array);
	  if (shift_ok == 0) return;

	  // default ghostsize is 4
	  //int ghostSize = 4;
	  int ghostSize = 0;
	    // check if we have defined SHADOW directive
	  Shadow_def * dsdef = get_defined_shadow(array);
	    // create a new shadow def to this symbol
          Shadow_def * sdef = new Shadow_def (DETECTED);
          for ( int iDim = 0; iDim < numDim; iDim++) {
		int lgSize=0,ugSize=0;
		int need_ghost=1;
		// try best to reduce the ghostarea .
		if (array_dim_collapsed(array,iDim) == 1) need_ghost=0;
		// stride not same
		if (rhsref->indexNumberOnDim(iDim) != -1) {
			int indexNumber = rhsref->indexNumberOnDim(iDim);
			int rDim = lhsref->dimNumberOnIndex(indexNumber);
			if (rDim != -1) {
				SgExpression *rstride = rhsref->getstride(iDim,indexNumber);
				SgExpression *lstride = lhsref->getstride(rDim,indexNumber);
				if (rstride->isInteger() && lstride->isInteger() ) 
				{
					if (rstride->valueInteger() != lstride->valueInteger())
						need_ghost=0;
				}
			}
		}
		if (need_ghost) {
		  if (dsdef) {
			lgSize = dsdef->getLower(iDim);
			ugSize = dsdef->getUpper(iDim);
		  }
		  else {
			lgSize = ugSize = ghostSize;
		  }
		}
/*
		if (need_ghost)  gSize = ghostSize;
		else gSize=0;
*/
	  	sdef->setUpper(iDim, lgSize);
	  	sdef->setLower(iDim, ugSize);	      
		  // this dim require SHIFT
		if (rhsref->indexNumberOnDim(iDim) == -1)
			ishift = 0;
          };
          sdef->setScope(s);
          array->addAttribute(Shad_Def,(void *) sdef, sizeof(Shadow_def));
//printf("set global array:%s,id:%d\n",array->identifier(),array->id());

      	  if (rhsref->hasScalarDim() && ishift)
       	  	  rhsref->setIShift();

	    // set Shadow_def
      	  rhsref->setShadow(sdef);
	}
}

// check remap once more , pick up the case that rhs has less dimension , such as B(4)=>a(2,4)
// set Amount
void set_remap_mapping(forall_reference * lhsref, forall_reference * rhsref)
{
	int i,j;

	switch (rhsref->getCommunication())
	{
	case UNKNOW:
	case REMAP:
	case BROADCAST_DIM:
	case BROADCAST_SCALAR:
  		SgArrayRefExp * lhsarray = (SgArrayRefExp *) lhsref->getexpression();
  		SgArrayRefExp * rhsarray = (SgArrayRefExp *) rhsref->getexpression();
		for (i=0; i<MAX_ALIGN_DIM; i++)
       	 	{
			int dim=rhsref->indexNumberOnDim(i);
			if (dim==-1) {
                		rhsref->setAmount(i, -1);
				continue;
			}
			for (j=0; j<lhsarray->numberOfSubscripts(); j++)
				if (lhsref->indexNumberOnDim(j)==dim)
					break;
			ASSERT(lhsref->indexNumberOnDim(j)==dim);
	    		rhsref->setAmount(i, j);
		}
		break;
	}
}

void outofcore_as_incore(forall_reference * lhsref, forall_reference * rhsref)
{
	SgExpression *lhsArray = lhsref->getexpression();
	SgSymbol *lhsSymb = lhsArray->symbol();
	SgExpression *rhsArray = rhsref->getexpression();
	SgSymbol *rhsSymb = rhsArray->symbol();

 	if (rhsref->getCommunication() == REMAP || rhsref->getCommunication() == UNKNOW)
	if (is_out_of_core(lhsSymb) && lhsref->numberOfIndex() == 1) 
	{
		lhsArray->addAttribute(SET_INCORE, NULL, 0);
		rhsArray->addAttribute(SET_INCORE, NULL, 0);
        }
	if (rhsref->getCommunication() == BROADCAST_DIM )
	if (is_out_of_core(lhsSymb) && rhsref->numberOfIndex() == 1) 
	{
		rhsArray->addAttribute(SET_INCORE, NULL, 0);
	}
}

void detect_communication(SgStatement * s, int counter)
{
  Forall_header_def * fhdef = (Forall_header_def *) s->attributeValue(0, FHDER_Def);

  SgStatement * body = s->childList1(0);

  forall_reference * lhsref = (forall_reference *)body->attributeValue(0, Forall_Ref);
  SgArrayRefExp * lhsarray = (SgArrayRefExp *) lhsref->getexpression();
  SgSymbol * lhssymb = lhsarray->symbol();

  forall_reference * rhsref = (forall_reference *)body->attributeValue(counter, Forall_Ref);
  SgArrayRefExp * rhsarray = (SgArrayRefExp *) rhsref->getexpression();
  SgSymbol * rhssymb = rhsarray->symbol();

  Alignment_def  *ladef = NULL, *radef = NULL;
  Distribute_def *lddef = NULL, *rddef = NULL;

  int ishift = 1;

  // LHS
  if (lhssymb->numberOfAttributes(Align_Def)!=0) {
    	ladef = (Alignment_def *)(lhssymb->attributeValue(0,Align_Def));
  	SgSymbol * lt = ladef->gettosymbol();
	if (lt->numberOfAttributes(Dist_Def) == 0) {
           	rhsref->setCommunication(UNKNOW);
		return;
	}
	lddef = (Distribute_def *)lt->attributeValue(0, Dist_Def);
  }
  else {
	if (lhssymb->numberOfAttributes(Dist_Def) == 0) {
	   // INHERIT
    	   if (lhssymb->numberOfAttributes(Inherit_Def)!=0)
           	rhsref->setCommunication(UNKNOW);
	   else	// default mapping (collapsed)
           	rhsref->setCommunication(REMAP);
           return;
	}
	else {
	   lddef = (Distribute_def *)lhssymb->attributeValue(0, Dist_Def);
	}
  }
  // RHS
  if (rhssymb->numberOfAttributes(Align_Def)!=0) {
      	radef = (Alignment_def *)(rhssymb->attributeValue(0,Align_Def));
	SgSymbol * rt = radef->gettosymbol();
	if (rt->numberOfAttributes(Dist_Def) == 0) {
           	rhsref->setCommunication(UNKNOW);
		return;
	}
	rddef = (Distribute_def *) rt->attributeValue(0, Dist_Def);
  }
  else {
	if (rhssymb->numberOfAttributes(Dist_Def) == 0) {
	   // INHERIT
    	   if (rhssymb->numberOfAttributes(Inherit_Def)!=0)
           	rhsref->setCommunication(UNKNOW);
	   else	// default mapping (collapsed)
           	rhsref->setCommunication(NONE);
           return;
	}
	else {
	   rddef = (Distribute_def *)rhssymb->attributeValue(0, Dist_Def);
	}
  }
      
  // try to decide if we can use shift, which is better than remap.

  if (lddef->gettosymbol()!=rddef->gettosymbol())  {
	printf("Error: Data aligned to different processors!\n");
	ASSERT(0);
  }

    // check if same subgroup
  if (same_processor_group(lhssymb, rhssymb) == 0) {
	rhsref->setCommunication(REMAP);
	return;
  }

  Processor_def *pdef = (Processor_def *)lddef->gettosymbol()->attributeValue(0, Proc_Def);

    //  check along dimension of processor
  for (int i=0; i< pdef->getDim(); i++) {

    	// i = 0; always one dim processor array.
    	//  if the partitioned array dim are both scalar or both in assign section,
    	//    do pcrc_detect_communication;
    	//  else 
    	//    do remap;

	int lpd=get_partitioned_dim(pdef,i, lhsarray);
	int rpd=get_partitioned_dim(pdef,i, rhsarray);
//printf("lpd=%d,rpd=%d\n",lpd,rpd);

// zheng.gb

	  // rhs is all collapsed
	if (rpd==-2) {
		rhsref->setCommunication(NONE);
		return;
	};

	  // if lhs is all collapsed ,but rhs is not, we should communicate.
	if (lpd==-2) {
		rhsref->setCommunication(REMAP);
		return;
	}

	  //  align a(i) with T(i,j)   ;  distribute T(block,block) onto P(2,2)
	  //  consider the second dim of P
	if (lpd == -1 || rpd == -1)
	{
		if (rpd == -1) {	// this dim is collapsed, check next one.
			continue;
		}
		else {
			rhsref->setCommunication(REMAP);
			return;
		}
	}

	int lindex = lhsref->indexNumberOnDim(lpd);
        int rindex = rhsref->indexNumberOnDim(rpd);

	  // cannot compare if the forall index is not same
	if (lindex != rindex) {
		rhsref->setCommunication(REMAP);
		return;
	}

	  // now the lpd and rpd contain same forall index or both scalar
	int amount;
	int lb, ub, stride;
	int ls, lo, rs, ro;
	detect_DIM leftDIM, rightDIM;

	if (fill_DIM(lhsref, lpd, leftDIM)==0) {	// something unknown, such as template size.
		  // hope to find a remap to settle down the communication type
		rhsref->setCommunication(UNKNOW);
		continue;
	}
	    
	if (fill_DIM(rhsref, rpd, rightDIM)==0)		// something unknown, such as template size.
	{
		rhsref->setCommunication(UNKNOW);
		continue;
	}

	if (lindex == -1) {		// both scalar, because lindex == rindex
		lb=0;   ub=0;
		stride=1;
		ls=1;   rs=1;
		SgExpression *loexp, *roexp;
		loexp = lhsarray->subscript(lpd);
		roexp = rhsarray->subscript(rpd);
		if ( !loexp->isInteger() || !roexp->isInteger() )
		{
		   rhsref->setCommunication(UNKNOW);
		   continue;
		}
		lo=loexp->valueInteger();
		ro=roexp->valueInteger();
	}
	else {
		SgExpression *lbexp,*ubexp,*strideexp;
		lbexp=fhdef->getindextriple(lindex, 0);
		ubexp=fhdef->getindextriple(lindex, 1);
		strideexp=fhdef->getindextriple(lindex, 2);
		if (!lbexp->isInteger() || !ubexp->isInteger() || (strideexp && !strideexp->isInteger()) )
		{
		   rhsref->setCommunication(UNKNOW);
		   continue;
		}
		lb = lbexp->valueInteger();
		ub = ubexp->valueInteger();
		if (strideexp) stride=strideexp->valueInteger();
		else stride=1;

		SgExpression *lsexp,*loexp;
		lsexp = lhsref->getstride(lpd, lindex);
		ASSERT(lsexp);
		loexp = lhsref->getoffset(lpd, lindex);
		if ( (lsexp && !lsexp->isInteger() ) || (loexp && !loexp->isInteger() ) )
		{
		   rhsref->setCommunication(UNKNOW);
		   continue;
		}
		if (lsexp) ls = lsexp->valueInteger();
		else  ls = 0;
		if (loexp) lo = loexp->valueInteger();
		else  lo = 0;

		SgExpression *rsexp,*roexp;
		rsexp = rhsref->getstride(rpd, rindex);
		ASSERT(rsexp);
		roexp = rhsref->getoffset(rpd, rindex);
		if ( (rsexp && !rsexp->isInteger() ) || (roexp && !roexp->isInteger() ) )
		{
		   rhsref->setCommunication(UNKNOW);
		   continue;
		}
		if (rsexp) rs = rsexp->valueInteger();
		else  rs = 0;
		if (roexp) ro = roexp->valueInteger();
		else  ro = 0;

	}
		 
//printf("lb:%d,ub:%d,stride:%d,ls:%d,lo:%d,rs:%d,ro:%d.\n",lb,ub,stride,ls,lo,rs,ro);
	  // now it's ready to detect communication 
	int detect_result = pcrc_detect_communication(&leftDIM,&rightDIM,&lb,&ub,&stride,&ls,&lo,&rs,&ro,&amount);
//printf("detect_result:%d\n",detect_result);
	switch (detect_result) {
	case 0:
	case 1:
		if (rhsref->getCommunication() != UNKNOW) 
		{
		  rhsref->setCommunication(SHIFT);
		  rhsref->setAmount(rpd, amount);
		}
		break;
	case 2:
		rhsref->setCommunication(REMAP);
		return;
		break;
	default:
		ASSERT(0);
	};		// end of switch

	if (lhsref->indexNumberOnDim(lpd) == -1 && rhsref->getCommunication() != NONE)
		ishift = 0;

  }   // end of for each dim

    // have tried best to exclude REMAP from UNKNOW
  if (rhsref->getCommunication() == UNKNOW) return;

    // check shift zero;
  if (rhsref->getCommunication()==SHIFT) {
    int move=0;
    for (int i=0; i<MAX_ALIGN_DIM; i++) {
      if (rhsref->getAmount(i)!=0) {
	move=1;
	break;
      }
    }
    if (move!=1)
      rhsref->setCommunication(NONE);
  }

  if (rhsref->getCommunication()==SHIFT && shift_is_ok(s,rhssymb)==0) {
printf("Reset %s as remap due to shift is not ok.\n",  rhssymb->identifier());
      rhsref->setCommunication(REMAP);
      return;
  }

    // check forall execution order;
  if (rhsref->getCommunication()==NONE || rhsref->getCommunication()==SHIFT) {
    if (forall_carry_dependent(fhdef,lhsref, rhsref)) {
      printf("because of flow dependence, %s's NONE or SHIFT is changed into REMAP.\n",rhssymb->identifier());
      rhsref->setCommunication(REMAP);
      return;
    }
  }

  // remap cyclic.
  if (rhsref->getCommunication()==SHIFT) {
    for (int i=0; i<MAX_ALIGN_DIM; i++) {
      if (rhsref->getAmount(i)!=0) {
	if (rddef->distCode(i)==CYCLIC) {
	  rhsref->setCommunication(REMAP);
printf(" reset as remap(cyclic), %d th dim amount %d\n",i,  rhsref->getAmount(i));
	  break;
	}
      }
    }
  }

  // Record template shadow  size
  if (rhsref->getCommunication()==SHIFT)  {


    Shadow_def * dsdef = get_defined_shadow(rhssymb);

    int ghostSize = 30;
 	    
    int shadow=1;

    for ( int dim =0; dim < MAX_ALIGN_DIM; dim++) {
      if (dsdef) {	// SHADOW directive
	int gSize = rhsref->getAmount(dim);
        if (gSize>0) {
		if (gSize > dsdef->getUpper(dim)) shadow=0;
	}
	else if (gSize<0) {
		if (-gSize > dsdef->getLower(dim)) shadow=0;
	     }
      }
      else if ( abs(rhsref->getAmount(dim)) > ghostSize ) {
	shadow=0;
	break;
       }
    }

    if (shadow == 0) rhsref->setCommunication(REMAP);

      // create Shadow_def(DETECTED)
    if (shadow) 
    {
      Shadow_def * sdef = new Shadow_def (DETECTED);
      
      for ( int iDim = 0; iDim < MAX_ALIGN_DIM; iDim++) {
	int gSize = rhsref->getAmount(iDim);
	if (gSize>0)
	  sdef->setUpper(iDim, abs(gSize));
	else
	  sdef->setLower(iDim, abs(gSize));	      
      };
      
      sdef->setScope(s);
      rhssymb->addAttribute(Shad_Def,(void *) sdef, sizeof(Shadow_def));

      if (rhsref->hasScalarDim() && ishift)
       rhsref->setIShift();
	// set Shadow_def
      rhsref->setShadow(sdef);
    }
  }

}

void detect_forall_reference(SgStatement * s)
{
  int header_evaluatable=1;

  Forall_header_def * fhdef = (Forall_header_def *) s->getAttribute(0, FHDER_Def)->getAttributeData();

  for (int i=0; i<fhdef->getindexnumber(); i++) 
    header_evaluatable *= fhdef->evaluatable_index(i);

  SgStatement * body=s->childList1(0);
  
  switch(body->variant()) {
  case ASSIGN_STAT: 
    {
      int i;

      for (i=0; i<body->numberOfAttributes(Forall_Ref); i++) {
	forall_reference *fref = (forall_reference *) body->attributeValue(i, Forall_Ref);

	if (!fref->checkRegular()) {
	  fref->setCommunication(IRREG);
	  if (!silentFlag)
	  printf("Irregular forall reference - no.%d!\n",i);
	}
      };
      
      forall_reference * lhsref = (forall_reference *) body->attributeValue(0, Forall_Ref);
      SgExpression *lhsArray = lhsref->getexpression();
      SgSymbol *lhsSymb = lhsArray->symbol();

      if (lhsref->getexpression()->variant()!=ARRAY_REF)
	printf("Warning: Multi assignment!\n");

      // check if the dim of Lhs is equal to the index number
      if (lhsref->numberOfDiffIndex() != fhdef->getindexnumber())
   	printf("Warning: In forall statement, write to Lhs more than twice.\n");

      // donot consider lhs
      for (i=1; i<body->numberOfAttributes(Forall_Ref); i++) {
	forall_reference *rhsref = (forall_reference *)body->attributeValue(i, Forall_Ref);
	SgExpression *rhsArray = rhsref->getexpression();
	SgSymbol *rhsSymb = rhsArray->symbol();

	if (rhsArray->variant()==ARRAY_REF) {

	    // already set to irregular
	  if (rhsref->getCommunication() != NONE) continue;

		// add by zheng.gb	1997/9/16
	    // if rhs is the same as lhs , none communication
	  if (*lhsref->getexpression() == *rhsref->getexpression())
	  {
		rhsref->setCommunication(NONE);
		continue;
	  }

	  if (is_resident(rhsSymb) == 1)
	  {
		rhsref->setCommunication(NONE);
	  }
	  else
	  if (array_all_collapsed(rhsSymb) == 1) 	// all collapsed
	  {
		rhsref->setCommunication(NONE);
	  }
	  else
          if (rhsref->isScalar())  
	  {
		rhsref->setCommunication(BROADCAST_SCALAR);
          }
	  else
	  if (rhsref->numberOfIndex() < lhsref->numberOfIndex() )
	  {
		rhsref->setCommunication(BROADCAST_DIM);
	  }
	  else
	  if (is_dynamic(rhsSymb) || is_dynamic(lhsSymb) ||
              is_pointer(rhsSymb) || is_pointer(lhsSymb) ||
              is_allocatable(rhsSymb) || is_allocatable(lhsSymb) )
	  {
		rhsref->setCommunication(UNKNOW);
	  }
	  else
		detect_communication(s, i);

	  // check forall execution order;
  	  // modified by slx 97/12/5
  	  if (rhsref->getCommunication()==NONE || rhsref->getCommunication()==SHIFT) 
	  {
    	   if (forall_carry_dependent(fhdef,lhsref, rhsref)) {
      		printf("because of flow dependence, %s's NONE or SHIFT is changed into REMAP.\n",rhsref->getexpression()->symbol()->identifier());
          	if (rhsref->isScalar()) 
			rhsref->setCommunication(BROADCAST_SCALAR);
	  	else if (rhsref->numberOfIndex() < lhsref->numberOfIndex() )
			rhsref->setCommunication(BROADCAST_DIM);
		else
	        	rhsref->setCommunication(REMAP);
    	   }
  	  }

	  // set remap mapping vector
	  set_remap_mapping(lhsref,rhsref);

 	  // set unknown ghostArea
	  set_unknow_ghostarea(s,rhsref);

  	  // see if can optimize out_of_core
	  outofcore_as_incore(lhsref, rhsref);

	} // end of if ARRAY_REF

      };  // end of for

      break;
    };
  case CONTROL_END: 
    break;    
  default:
    printf("Not touched yet!!!\n");
    break;
  };
}

void show_forall_reference(SgStatement * s)
{
  s=s->childList1(0);

  switch(s->variant()) {
  case ASSIGN_STAT: 
    {
      int i;
      for (i=0; i<s->numberOfAttributes(Forall_Ref); i++) {
	((forall_reference *) ((s->getAttribute(i, Forall_Ref))->getAttributeData())) ->show();
      };
    };
    break;

  case CONTROL_END: 
    break;
    
  default:
    printf("Not touched yet!!\n");
    break;
  };
}

void index_forall_reference(int indexdim, SgSymbol * index, SgStatement * s)
{
  switch(s->variant()) {
  case ASSIGN_STAT:     {
    for (int i=0; i<s->numberOfAttributes(Forall_Ref); i++) {
      forall_reference * fref= (forall_reference *) s->attributeValue(i, Forall_Ref);
      
      int indexlevel;
      
      SgExpression * frexp=fref->getexpression();
      
      switch(frexp->variant()) {
      case VAR_REF:
	indexlevel = getreflevel(index, frexp);
	fref->setIndex(0, indexdim, indexlevel);	  
	break;
      case ARRAY_REF:	  {
	SgArrayRefExp * arrayexp=(SgArrayRefExp *) frexp;
	
	SgSymbol * arraysym=arrayexp->symbol();
	
	for (int dimI=0; dimI<arrayexp->numberOfSubscripts();dimI++) {
	  indexlevel=getreflevel(index, arrayexp->subscript(dimI));
	  fref->setIndex(dimI, indexdim, indexlevel);
	  
	  // if possible, try to get stride and offset;
	  
	  if (indexlevel==1) {
/*	have bugs in getlinerexp
	    SgExpression * stride = getlinerexp(1, index, arrayexp->subscript(dimI));
	    SgExpression * offset = getlinerexp(0, index, arrayexp->subscript(dimI));
*/
	    SgExpression *exp=coefficient(arrayexp->subscript(dimI), index);
	    SgExpression * stride = exp->lhs();
	    SgExpression * offset = exp->rhs();
	    stride = calculateExp(stride);
	    offset = calculateExp(offset);
	    
	    fref->setstride(dimI, indexdim, stride);
	    fref->setoffset(dimI, indexdim, offset);
	  }
	}
	
	break;
      };
      
      default:
	printf("index_forall_reference::Expression not touched!\n");
	break;
      };
    };
    break;
  };

  case CONTROL_END: 
    break;
    
  default:
    printf("Statement not touched yet!!!\n");
    break;
  };
}

//---------------------

void add_triple(SgStatement *s, SgExpression * e, Forall_header_def * fhdef)
{
  SgExpression *st=e->operand(2);
  if (st != NULL)
    fhdef->setindextriple(2, st);
  SgExpression *l =  e->operand(1)->operand(1);
  fhdef->setindextriple(0, l);
  SgExpression *u =  e->operand(1)->operand(2);
  fhdef->setindextriple(1, u);
}

void add_index(SgStatement *s, SgExpression * e, Forall_header_def * fhdef)
{
  //printf("The id is %d\n", e->id());
  fhdef->setindexsymb(e->operand(1)->operand(1)->symbol());
  add_triple(s, e->operand(1)->operand(2), fhdef);

  if ((e->variant()==CONTROL_LIST) && (e->operand(2)!=NULL)) 
    add_index(s, e->operand(2), fhdef);
}

void add_forall_index(SgStatement * s, SgSymbol * symb)
{
  Forall_index_def * fidef = new Forall_index_def;
  fidef->setsymbol(symb);

  // add this attributes to both program header and symbol

  s->controlParent()->addAttribute(FINDEX_Def,(void *) fidef, sizeof(Forall_index_def));
  symb->addAttribute(FINDEX_Def,(void *) fidef, sizeof(Forall_index_def));

}

void analyse_forall_header(SgStatement * s)
{
  Forall_header_def * fhdef = new Forall_header_def;

  if (s->expr(1)!=NULL)
    fhdef->setcondition(s->expr(1));

  add_index(s, s->expr(0)->operand(1), fhdef);

  // add this attributes to forall statement
  s->addAttribute(FHDER_Def,(void *) fhdef, sizeof(Forall_header_def));

  // add this attributes to program header;
  s->controlParent()->addAttribute(FHDER_Def,(void *) fhdef, sizeof(Forall_header_def));

  for (int i=0; i<fhdef->getindexnumber(); i++) {
    SgSymbol * index=fhdef->getindexsymb(i);
      
    if(index->numberOfAttributes(FINDEX_Def)==0)
      add_forall_index(s, index);
  }

    // analyse expression of header
}

